# -*- coding: utf-8 -*-

"""

keras_resnet.blocks._time_distributed_2d
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements a number of popular time distributed two-dimensional residual blocks.

"""

import keras.layers
import keras.regularizers

parameters = {
    "kernel_initializer": "he_normal"
}


def time_distributed_basic_2d(filters, strides=(1, 1), first=False):
    """

    A time distributed two-dimensional basic block.

    :param filters: the output’s feature space

    :param strides: the convolution’s stride

    :param first: whether this is the first instance inside a residual blocks

    Usage:

        >>> import keras_resnet.blocks

        >>> keras_resnet.blocks.time_distributed_basic_2d(64)

    """
    def f(x):
        if keras.backend.image_data_format() == "channels_last":
            axis = 3
        else:
            axis = 1

        y = keras.layers.TimeDistributed(keras.layers.Conv2D(filters, (3, 3), strides=strides, padding="same", **parameters))(x)

        y = keras.layers.TimeDistributed(keras.layers.BatchNormalization(axis=axis))(y)
        y = keras.layers.TimeDistributed(keras.layers.Activation("relu"))(y)

        y = keras.layers.TimeDistributed(keras.layers.Conv2D(filters, (3, 3), padding="same", **parameters))(y)

        y = keras.layers.TimeDistributed(keras.layers.BatchNormalization(axis=axis))(y)
        y = _shortcut(x, y)
        y = keras.layers.TimeDistributed(keras.layers.Activation("relu"))(y)

        return y

    return f


def time_distributed_bottleneck_2d(filters, strides=(1, 1), first=False):
    """

    A time distributed bottleneck two-dimensional block.

    :param filters: the output’s feature space

    :param strides: the convolution’s stride

    :param first: whether this is the first instance inside a residual blocks

    Usage:

        >>> import keras_resnet.blocks

        >>> keras_resnet.blocks.time_distributed_bottleneck_2d(64)

    """
    def f(x):
        if keras.backend.image_data_format() == "channels_last":
            axis = 3
        else:
            axis = 1

        if first:
            y = keras.layers.TimeDistributed(keras.layers.Conv2D(filters, (1, 1), strides=strides, padding="same", **parameters))(x)
        else:
            y = keras.layers.TimeDistributed(keras.layers.Conv2D(filters, (3, 3), strides=strides, padding="same", **parameters))(x)

        y = keras.layers.TimeDistributed(keras.layers.BatchNormalization(axis=axis))(y)
        y = keras.layers.TimeDistributed(keras.layers.Activation("relu"))(y)

        y = keras.layers.TimeDistributed(keras.layers.Conv2D(filters, (3, 3), padding="same", **parameters))(y)

        y = keras.layers.TimeDistributed(keras.layers.BatchNormalization(axis=axis))(y)
        y = keras.layers.TimeDistributed(keras.layers.Activation("relu"))(y)

        y = keras.layers.TimeDistributed(keras.layers.Conv2D(filters * 4, (1, 1), **parameters))(y)

        y = keras.layers.TimeDistributed(keras.layers.BatchNormalization(axis=axis))(y)
        y = _shortcut(x, y)
        y = keras.layers.TimeDistributed(keras.layers.Activation("relu"))(y)

        return y

    return f


def _shortcut(a, b):
    a_shape = keras.backend.int_shape(a)[1:]
    b_shape = keras.backend.int_shape(b)[1:]

    if keras.backend.image_data_format() == "channels_last":
        x = int(round(a_shape[1] / b_shape[1]))
        y = int(round(a_shape[2] / b_shape[2]))

        if x > 1 or y > 1 or not a_shape[3] == b_shape[3]:
            a = keras.layers.TimeDistributed(keras.layers.Conv2D(b_shape[3], (1, 1), strides=(x, y), padding="same", **parameters))(a)

            a = keras.layers.TimeDistributed(keras.layers.BatchNormalization(axis=3))(a)
    else:
        x = int(round(a_shape[2] / b_shape[2]))
        y = int(round(a_shape[3] / b_shape[3]))

        if x > 1 or y > 1 or not a_shape[1] == b_shape[1]:
            a = keras.layers.TimeDistributed(keras.layers.Conv2D(b_shape[1], (1, 1), strides=(x, y), padding="same", **parameters))(a)

            a = keras.layers.TimeDistributed(keras.layers.BatchNormalization(axis=1))(a)

    return keras.layers.add([a, b])



# -*- coding: utf-8 -*-

"""

keras_resnet.models._2d
~~~~~~~~~~~~~~~~~~~~~~~

This module implements popular two-dimensional residual models.

"""

import keras.backend
import keras.layers
import keras.models
import keras.regularizers

import keras_resnet.blocks


class ResNet(keras.models.Model):
    """

    A custom :class:`ResNet <ResNet>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param block: a residual blocks (e.g. an instance of `keras_resnet.basic`)

    Usage:

        >>> import keras_resnet.blocks
        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> blocks = [2, 2, 2, 2]

        >>> block = keras_resnet.blocks.basic_2d

        >>> y = keras_resnet.models.ResNet(x, classes, blocks, block)

        >>> y = keras.layers.Flatten()(y.output)

        >>> y = keras.layers.Dense(classes, activation="softmax")(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])

    """

    def __init__(self, inputs, blocks, block):
        if keras.backend.image_data_format() == "channels_last":
            axis = 3
        else:
            axis = 1

        x = keras.layers.Conv2D(64, (7, 7), strides=(2, 2), padding="same")(inputs)
        x = keras.layers.BatchNormalization(axis=axis)(x)
        x = keras.layers.Activation("relu")(x)
        x = keras.layers.MaxPooling2D((3, 3), strides=(2, 2), padding="same")(x)

        features = 64

        for j, iterations in enumerate(blocks):
            for k in range(iterations):
                if k == 0 and not j == 0:
                    strides = (2, 2)
                else:
                    strides = (1, 1)

                x = block(features, strides, j == 0 and k == 0)(x)

            features *= 2

        x = keras.layers.BatchNormalization(axis=axis)(x)
        x = keras.layers.Activation("relu")(x)

        shape = keras.backend.int_shape(x)

        if keras.backend.image_data_format() == "channels_last":
            pool_size = (shape[1], shape[2])
        else:
            pool_size = (shape[2], shape[3])

        x = keras.layers.AveragePooling2D(pool_size, strides=(1, 1))(x)

        super(ResNet, self).__init__(inputs, x)


class ResNet18(ResNet):
    """

    A :class:`ResNet18 <ResNet18>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.ResNet18(x)

        >>> y = keras.layers.Flatten()(y.output)

        >>> y = keras.layers.Dense(classes, activation="softmax")(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])

    """

    def __init__(self, inputs):
        blocks = [2, 2, 2, 2]

        block = keras_resnet.blocks.basic_2d

        super(ResNet18, self).__init__(inputs, blocks, block)


class ResNet34(ResNet):
    """

    A :class:`ResNet34 <ResNet34>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.ResNet34(x)

        >>> y = keras.layers.Flatten()(y.output)

        >>> y = keras.layers.Dense(classes, activation="softmax")(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])

    """

    def __init__(self, inputs):
        blocks = [3, 4, 6, 3]

        block = keras_resnet.blocks.basic_2d

        super(ResNet34, self).__init__(inputs, blocks, block)


class ResNet50(ResNet):
    """

    A :class:`ResNet50 <ResNet50>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.ResNet50(x)

        >>> y = keras.layers.Flatten()(y.output)

        >>> y = keras.layers.Dense(classes, activation="softmax")(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])

    """

    def __init__(self, inputs):
        blocks = [3, 4, 6, 3]

        block = keras_resnet.blocks.bottleneck_2d

        super(ResNet50, self).__init__(inputs, blocks, block)


class ResNet101(ResNet):
    """

    A :class:`ResNet101 <ResNet101>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.ResNet101(x)

        >>> y = keras.layers.Flatten()(y.output)

        >>> y = keras.layers.Dense(classes, activation="softmax")(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])

    """

    def __init__(self, inputs):
        blocks = [3, 4, 23, 3]

        block = keras_resnet.blocks.bottleneck_2d

        super(ResNet101, self).__init__(inputs, blocks, block)


class ResNet152(ResNet):
    """

    A :class:`ResNet152 <ResNet152>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.ResNet152(x)

        >>> y = keras.layers.Flatten()(y.output)

        >>> y = keras.layers.Dense(classes, activation="softmax")(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])

    """

    def __init__(self, inputs):
        blocks = [3, 8, 36, 3]

        block = keras_resnet.blocks.bottleneck_2d

        super(ResNet152, self).__init__(inputs, blocks, block)


class ResNet200(ResNet):
    """

    A :class:`ResNet200 <ResNet200>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.ResNet200(x)

        >>> y = keras.layers.Flatten()(y.output)

        >>> y = keras.layers.Dense(classes, activation="softmax")(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])

    """

    def __init__(self, inputs):
        blocks = [3, 24, 36, 3]

        block = keras_resnet.blocks.bottleneck_2d

        super(ResNet200, self).__init__(inputs, blocks, block)



# -*- coding: utf-8 -*-

"""

keras_resnet.models._time_distributed_2d
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements popular time distributed two-dimensional residual networks.

"""

import keras.backend
import keras.layers
import keras.models
import keras.regularizers

import keras_resnet.blocks


class TimeDistributedResNet(keras.models.Model):
    """
    
    A custom :class:`TimeDistributedResNet <TimeDistributedResNet>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)
    
    :param blocks: the network’s residual architecture

    :param block: a time distributed residual block (e.g. an instance of `keras_resnet.blocks.time_distributed_basic_2d`)

    Usage:

        >>> import keras_resnet.blocks
        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> blocks = [2, 2, 2, 2]

        >>> blocks = keras_resnet.blocks.time_distributed_basic_2d

        >>> y = keras_resnet.models.TimeDistributedResNet(x, classes, blocks, blocks)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    
    """

    def __init__(self, inputs, blocks, block):
        if keras.backend.image_data_format() == "channels_last":
            axis = 3
        else:
            axis = 1

        x = keras.layers.TimeDistributed(keras.layers.Conv2D(64, (7, 7), strides=(2, 2), padding="same"))(inputs)
        x = keras.layers.TimeDistributed(keras.layers.BatchNormalization(axis=axis))(x)
        x = keras.layers.TimeDistributed(keras.layers.Activation("relu"))(x)
        x = keras.layers.TimeDistributed(keras.layers.MaxPooling2D((3, 3), strides=(2, 2), padding="same"))(x)

        features = 64

        for j, iterations in enumerate(blocks):
            for k in range(iterations):
                if k == 0 and not j == 0:
                    strides = (2, 2)
                else:
                    strides = (1, 1)

                x = block(features, strides, j == 0 and k == 0)(x)

            features *= 2

        x = keras.layers.TimeDistributed(keras.layers.BatchNormalization(axis=axis))(x)
        x = keras.layers.TimeDistributed(keras.layers.Activation("relu"))(x)

        shape = keras.backend.int_shape(x)

        if keras.backend.image_data_format() == "channels_last":
            pool_size = (shape[1], shape[2])
        else:
            pool_size = (shape[2], shape[3])

        x = keras.layers.TimeDistributed(keras.layers.AveragePooling2D(pool_size, strides=(1, 1)))(x)

        super(TimeDistributedResNet, self).__init__(inputs, x)


class TimeDistributedResNet18(TimeDistributedResNet):
    """
    
    A :class:`TimeDistributedResNet18 <TimeDistributedResNet18>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet18(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    
    """

    def __init__(self, inputs):
        blocks = [2, 2, 2, 2]

        block = keras_resnet.blocks.time_distributed_basic_2d

        super(TimeDistributedResNet18, self).__init__(inputs, blocks, block)


class TimeDistributedResNet34(TimeDistributedResNet):
    """
    
    A :class:`TimeDistributedResNet34 <TimeDistributedResNet34>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet34(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    
    """

    def __init__(self, inputs):
        blocks = [3, 4, 6, 3]

        block = keras_resnet.blocks.time_distributed_basic_2d

        super(TimeDistributedResNet34, self).__init__(inputs, blocks, block)


class TimeDistributedResNet50(TimeDistributedResNet):
    """
    
    A :class:`TimeDistributedResNet50 <TimeDistributedResNet50>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet50(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    
    """

    def __init__(self, inputs):
        blocks = [3, 4, 6, 3]

        block = keras_resnet.blocks.time_distributed_bottleneck_2d

        super(TimeDistributedResNet50, self).__init__(inputs, blocks, block)


class TimeDistributedResNet101(TimeDistributedResNet):
    """
    
    A :class:`TimeDistributedResNet101 <TimeDistributedResNet101>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet101(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    
    """

    def __init__(self, inputs):
        blocks = [3, 4, 23, 3]

        block = keras_resnet.blocks.time_distributed_bottleneck_2d

        super(TimeDistributedResNet101, self).__init__(inputs, blocks, block)


class TimeDistributedResNet152(TimeDistributedResNet):
    """
    
    A :class:`TimeDistributedResNet152 <TimeDistributedResNet152>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet152(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    
    """

    def __init__(self, inputs):
        blocks = [3, 8, 36, 3]

        block = keras_resnet.blocks.time_distributed_bottleneck_2d

        super(TimeDistributedResNet152, self).__init__(inputs, blocks, block)


class TimeDistributedResNet200(TimeDistributedResNet):
    """
    
    A :class:`TimeDistributedResNet200 <TimeDistributedResNet200>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet200(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    
    """

    def __init__(self, inputs):
        blocks = [3, 24, 36, 3]

        block = keras_resnet.blocks.time_distributed_bottleneck_2d

        super(TimeDistributedResNet200, self).__init__(inputs, blocks, block)
